# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2019 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import aiohttp
import asyncio
import logging
from typing import Union

from .cooldowns import RateBucket
from .errors import HTTPException, Unauthorized


log = logging.getLogger(__name__)


class HTTPSession:
    BASE = 'https://api.twitch.tv/helix'

    def __init__(self, loop, **attrs):
        self.client_id = client_id = attrs.get('client_id', None)

        if not client_id:
            log.warning('Running without client ID, some HTTP endpoints may not work without authentication.')

        self._bucket = RateBucket(method='http')
        self._session = aiohttp.ClientSession(loop=loop)

    async def request(self, method, url, *, params=None, limit=None, **kwargs):
        count = kwargs.pop('count', False)

        data = []

        params = params or []
        url = f'{self.BASE}{url}'

        # headers = {}

        headers = kwargs.pop('headers', {})

        if self.client_id is not None:
            headers['Client-ID'] = str(self.client_id)

        cursor = None

        def reached_limit():
            return limit and len(data) >= limit

        def get_limit():
            if limit is None:
                return '100'

            to_get = limit - len(data)
            return str(to_get) if to_get < 100 else '100'

        while not reached_limit():
            if cursor is not None:
                params.append(('after', cursor))

            params.append(('first', get_limit()))

            body, is_text = await self._request(method, url, params=params, headers=headers, **kwargs)

            if is_text:
                return body

            if count:
                return body['total']

            params.pop()  # remove the first param

            if cursor is not None:
                params.pop()

            data += body['data']

            try:
                cursor = body['pagination'].get('cursor', None)
            except KeyError:
                break
            else:
                if not cursor:
                    break

        return data

    async def _request(self, method, url, utilize_bucket=True, **kwargs):
        reason = None

        for attempt in range(5):
            if utilize_bucket and self._bucket.limited:
                await self._bucket.wait_reset()

            async with self._session.request(method, url, **kwargs) as resp:
                if 500 <= resp.status <= 504:
                    reason = resp.reason
                    await asyncio.sleep(2 ** attempt + 1)
                    continue

                if utilize_bucket:
                    reset = resp.headers.get('Ratelimit-Reset')
                    remaining = resp.headers.get('Ratelimit-Remaining')

                    self._bucket.update(reset=reset, remaining=remaining)

                if 200 <= resp.status < 300:
                    if resp.content_type == 'application/json':
                        return await resp.json(), False

                    return await resp.text(encoding='utf-8'), True

                if resp.status == 401:
                    if self.client_id is None:
                        raise Unauthorized('A client ID or other authorization is needed to use this route.')

                    raise Unauthorized('You\'re not authorized to use this route.')

                if resp.status == 429:
                    reason = 'Ratelimit Reached'

                    if not utilize_bucket:  # non Helix APIs don't have ratelimit headers
                        await asyncio.sleep(3 ** attempt + 1)
                    continue

                raise HTTPException(f'Failed to fulfil request ({resp.status}).', resp.reason)

        raise HTTPException('Failed to reach Twitch API', reason)

    @staticmethod
    def _populate_entries(*channels: Union[str, int]):
        names = set()
        ids = set()

        for channel in channels:
            if isinstance(channel, str):
                if channel.isdigit():
                    # Handle ids in the string form
                    ids.add(int(channel))
                else:
                    names.add(channel)
            elif isinstance(channel, int):
                ids.add(str(channel))

        if len(names | ids) > 100:
            raise HTTPException('Bad Request - Total entries must not exceed 100.')

        return names, ids

    async def get_users(self, *users: Union[str, int]):
        names, ids = self._populate_entries(*users)
        params = [('id', x) for x in ids] + [('login', x) for x in names]

        return await self.request('GET', '/users', params=params)

    async def get_follow(self, from_id: str, to_id: str):
        params = [('from_id', from_id), ('to_id', to_id)]
        return await self.request('GET', '/users/follows', params=params)

    async def get_followers(self, user_id: str, *, count):
        params = [('to_id', user_id)]
        return await self.request('GET', '/users/follows', params=params, count=count)

    async def get_following(self, user_id: str, *, count):
        params = [('from_id', user_id)]
        return await self.request('GET', '/users/follows', params=params, count=count)

    async def get_streams(self, *, game_id=None, language=None, channels, limit=None):
        if channels:
            names, ids = self._populate_entries(*channels)
            params = [('user_id', x) for x in ids] + [('user_login', x) for x in names]
        else:
            params = []

        if game_id is not None:
            params.append(('game_id', str(game_id)))

        if language is not None:
            params.append(('language', language))

        return await self.request('GET', '/streams', params=params, limit=limit)

    async def get_games(self, *games: Union[str, int]):
        names, ids = self._populate_entries(*games)
        params = [('id', x) for x in ids] + [('name', x) for x in names]

        return await self.request('GET', '/games', params=params)

    async def get_top_games(self, limit=None):
        return await self.request('GET', '/games/top', limit=limit)

    async def modify_webhook_subscription(self, *, callback, mode, topic, lease_seconds, secret=None):
        data = {
            'hub.callback': callback,
            'hub.mode': mode,
            'hub.topic': topic,
            'hub.lease_seconds': lease_seconds
        }

        if secret is not None:
            data['secret'] = secret

        return await self.request('POST', '/webhooks/hub', json=data)

    async def create_clip(self, token: str, broadcaster_id: int):
        params = [('broadcaster_id', str(broadcaster_id))]
        return await self.request('POST', '/clips', params=params, headers={'Authorization': f'Bearer {token}'})



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2019 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

__all__ = (
    'Topic',
    'UserFollows',
    'StreamChanged',
    'UserChanged',
    'GameAnalytics',
    'ExtensionAnalytics',
    'TwitchWebhookServer',
)


import abc
import asyncio
import json
import uuid

from aiohttp import web

from twitchio.errors import HTTPException


class TwitchWebhookServer:

    def __init__(self, *, bot, local: str, external: str, port: int, callback: str=None):
        self._bot = bot
        self.local = local
        self.external = external
        self.port = port
        self.callback = callback or uuid.uuid4().hex
        self.app = web.Application()
        self.app.add_routes([web.get(f'/{self.callback}', self.handle_callback),
                             web.post(f'/{self.callback}', self.handle_callback_post)])

        self.loop = None

    def stop(self):
        self.loop.stop()

    def run_server(self, loop: asyncio.BaseEventLoop):
        asyncio.set_event_loop(loop)
        self.loop = loop

        handler = self.app.make_handler()
        server = loop.create_server(handler, host=self.local, port=self.port)

        loop.run_until_complete(server)
        loop.run_forever()

    async def handle_callback_post(self, request) -> web.Response:
        try:
            data = await request.json()
        except json.JSONDecodeError:
            return web.Response(text='Bad Request', status=400)

        asyncio.run_coroutine_threadsafe(self._bot.event_webhook(data), loop=self.loop)

        return web.Response(text='200: OK', status=200)

    async def handle_callback(self, request) -> web.Response:
        query = request.query

        try:
            if query['hub.mode'] == 'denied':
                asyncio.run_coroutine_threadsafe(self._bot._ws.event_error(
                    HTTPException(f'Webhook subscription denied | {query["hub.reason"]}')), loop=self.loop)
                return web.Response(text='200: OK', status=200)

            if query['hub.challenge']:
                asyncio.run_coroutine_threadsafe(self._bot.event_webhook(query), loop=self.loop)
                return web.Response(body=query['hub.challenge'],
                                    content_type='application/json')
        except KeyError:
            web.Response(text='Bad Request', status=400)

        return web.Response(text='200: OK', status=200)


class Topic(abc.ABC):
    """
    Represents a Topic which can be used to modify a Webhook subscription.

    .. note::
        You can not create an instance of this class, use one of the derived ones instead.
    """

    URL = None

    __slots__ = ()

    @property
    def _parameters(self):
        return [x for x in self.__slots__ if getattr(self, x) is not None]

    def as_uri(self):
        """
        Converts the Topic into the URI which can be used to create a Webhook subscription.

        Returns
        -------
        str
            The Topic as an URI.
        """

        params = '&'.join(f'{name}={getattr(self, name)}' for name in self._parameters)
        return f'{self.URL}?{params}'


class UserFollows(Topic):
    """
    This Topic notifies you whenever a user follows someone or someone is being followed.

    Parameters
    ----------
    first: Optional[int]
        This needs to be 1. Please see the Twitch documentation for more information.
    from_id: Optional[int]
        User ID for the user that's following other Twitch channels.
    to_id: Optional[int]
        User ID for the channel that other Twitch users follow.

    Raises
    ------
    TypeError
        Neither the `from_id` or `to_id` parameter were given.
    """

    URL = 'https://api.twitch.tv/helix/users/follows'

    __slots__ = ('first', 'from_id', 'to_id')

    def __init__(self, *, first=1, from_id=None, to_id=None):
        self.first = first

        if from_id is None and to_id is None:
            raise TypeError('Missing either "from_id" or "to_id" argument.')

        self.from_id = from_id
        self.to_id = to_id


class StreamChanged(Topic):
    """
    This Topic notifies you whenever a Stream starts, is modified or stops.

    Parameters
    ----------
    user_id: int
        The channel to receive notifications for.
    """

    URL = 'https://api.twitch.tv/helix/streams'

    __slots__ = ('user_id',)

    def __init__(self, user_id):
        self.user_id = user_id


class UserChanged(Topic):
    """
    This Topic notifies you whenever a user updates their profile.

    Parameters
    ----------
    user_id: int
        The user to receive information for.
    """

    URL = 'https://api.twitch.tv/helix/users'

    __slots__ = ('id',)

    def __init__(self, user_id):
        self.id = user_id


class GameAnalytics(Topic):
    """
    This Topic notifies you whenever a new game analytics report is available.

    .. note::
        This Topic requires the `analytics:read:games` OAuth scope.

    Parameters
    ----------
    game_id: int
        The game to receive notifications for.
    """

    URL = 'https://api.twitch.tv/helix/analytics/games'

    __slots__ = ('game_id',)

    def __init__(self, game_id):
        self.game_id = game_id


class ExtensionAnalytics(Topic):
    """
    This Topic notifies you whenever a new extension analytics report is available.

    ..note ::
        This Topic requires the `analytics:read:extensions` OAuth scope.

    Parameters
    ----------
    extension_id: int
        The extension to receive notifications for.
    """

    URL = 'https://api.twitch.tv/helix/analytics/extensions'

    __slots__ = ('extension_id',)

    def __init__(self, extension_id):
        self.extension_id = extension_id



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2019 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
from collections import namedtuple
from typing import Union

from twitchio.errors import HTTPException
from twitchio.http import HTTPSession


User = namedtuple('User', ('id', 'login', 'display_name', 'type', 'broadcaster_type', 'description',
                           'profile_image', 'offline_image', 'view_count'))
Chatters = namedtuple('Chatters', ('count', 'all', 'broadcaster', 'vips', 'moderators', 'staff',
                                   'admins', 'global_mods', 'viewers'))


class Client:

    def __init__(self, *, loop=None, client_id=None, **kwargs):
        loop = loop or asyncio.get_event_loop()
        self.http = HTTPSession(loop=loop, client_id=client_id)

    async def get_users(self, *users: Union[str, int]) -> list:
        """|coro|

        Method which retrieves user information on the specified names/ids.

        Parameters
        ------------
        \*users: str
            The user name(s)/id(s) to retrieve data for.

        Returns
        ---------
        list: User
            List containing user data in a namedtuple.

        Raises
        --------
        HTTPException
            Bad request while fetching stream.

        Notes
        -------
        .. note::
            This method accepts both user ids or names, or a combination of both. Multiple names/ids may be passed.
        """

        data = await self.http.get_users(*users)

        return [User(*user.values()) for user in data]

    async def get_stream(self, channel: Union[int, str]):
        """|coro|

        Method which retrieves stream information on the channel, provided it is active (Live).

        Parameters
        ------------
        channel: Union[int, str]
            The channel id or name to retrieve data for.

        Returns
        ---------
        dict
            Dict containing active streamer data. Could be None if the stream is not live.

        Raises
        --------
        HTTPException
            Bad request while fetching stream.
        """

        data = await self.http.get_streams(channels=[channel])

        try:
            return data[0]
        except IndexError:
            pass

    async def get_streams(self, *, game_id=None, language=None, channels=None, limit=None):
        """|coro|

        Method which retrieves multiple stream information on the given channels, provided they are active (Live).

        Parameters
        ------------
        game_id: Optional[int]
            The game to filter streams for.
        language: Optional[str]
            The language to filter streams for.
        channels: Optional[Union[int, str]]
            The channels in id or name form, to retrieve information for.
        limit: Optional[int]
            Maximum number of results to return.

        Returns
        ---------
        list
            List containing active streamer data. Could be None if none of the streams are live.

        Raises
        --------
        HTTPException
            Bad request while fetching streams.
        """

        return await self.http.get_streams(game_id=game_id, language=language, channels=channels, limit=limit)

    async def get_games(self, *games: Union[str, int]):
        """|coro|

        Method which retrieves games information on the given game ID(s)/Name(s).

        Parameters
        ------------
        \*games: Union[str, int]
            The games in either id or name form to retrieve information for.

        Returns
        ---------
        list
            List containing game information. Could be None if no games matched.

        Raises
        --------
        HTTPException
            Bad request while fetching games.
        """

        return await self.http.get_games(*games)

    async def get_top_games(self, limit=None):
        """|coro|

        Retrieves the top games currently being played on Twitch.

        Parameters
        ------------
        limit: Optional[int]
            Maximum amount of results to fetch.

        Returns
        ---------
        list
            List containing game information. Could be None if no games matched.

        Raises
        --------
        HTTPException
            Bad request while fetching games.
        """

        return await self.http.get_top_games(limit=limit)

    async def modify_webhook_subscription(self, *, callback, mode, topic, lease_seconds=0, secret=None):
        """|coro|

        Creates a webhook subscription.

        Parameters
        ----------
        callback: str
            The URL which will be called to verify the subscripton and on callback.
        mode: :class:`.WebhookMode`
            Mode which describes whether the subscription should be created or not.
        topic: :class:`.Topic`
            Details about the subscription.
        lease_seconds: Optional[int]
            How many seconds the subscription should last. Defaults to 0, maximum is 846000.
        secret: Optional[str]
            A secret string which Twitch will use to add the `X-Hub-Signature` header to webhook requests.
            You can use this to verify the POST request came from Twitch using `sha256(secret, body)`.

        Raises
        --------
        HTTPException
            Bad request while modifying the subscription.
        """

        await self.http.modify_webhook_subscription(
            callback=callback,
            mode=mode.name,
            topic=topic.as_uri(),
            lease_seconds=lease_seconds,
            secret=secret,
        )

    async def get_follow(self, from_id: Union[int, str], to_id: Union[int, str]):
        """|coro|

        Retrieves the follow relationship between two users, provided it exists.

        Parameters
        ------------
        from_id: Union[int, str]
            The user who is following.
        to_id: Union[int, str]
            The user being followed.

        Returns
        ---------
        dict
            Dict containing follow relationship data. Could be None if there is no follow relationship.

        Raises
        --------
        HTTPException
            Bad request while fetching follow relationship.
        """

        data = await self.http.get_follow(str(from_id), str(to_id))

        try:
            return data[0]
        except IndexError:
            pass

    async def get_followers(self, user_id: Union[int, str], *, count: bool=False):
        """|coro|

        Retrieves the list of users who are following a user.

        Parameters
        ------------
        user_id: Union[int, str]
            The user to retrieve the list of followers for.
        count: bool
            Whether to return only the total count of followers or not. Defaults to False.

        Returns
        ---------
        list
            List containing users following this user.
        int
            An int of the total count of followers. Only available when ``count``=True

        Raises
        --------
        HTTPException
            Bad request while fetching users.
        """

        return await self.http.get_followers(str(user_id), count=count)

    async def get_following(self, user_id: Union[int, str], count: bool=False):
        """|coro|

        Retrieves the list of users who this user is following.

        Parameters
        ------------
        user_id: Union[int, str]
            The user to retrieve the list of followed users for.
        count: bool
            Whether to return only the total count of followed users or not. Defaults to False.

        Returns
        ---------
        list
            List containing users that the user is following.
        int
            An int of the total count of followed users. Only available when ``count``=True

        Raises
        --------
        HTTPException
            Bad request while fetching users.
        """

        return await self.http.get_following(str(user_id), count=count)

    async def get_chatters(self, channel: str):
        """|coro|

        Method which retrieves the currently active chatters on the given stream.

        Parameters
        ------------
        channel: str
            The channel name to retrieve data for.

        Returns
        ---------
        Chatters
            Namedtuple containing active chatter data.

        Raises
        --------
        HTTPException
            Bad request while fetching stream chatters.
        """

        url = f'http://tmi.twitch.tv/group/user/{channel.lower()}/chatters'

        async with self.http._session.get(url) as resp:
            if 200 <= resp.status < 300:
                data = await resp.json()
            else:
                raise HTTPException(f'Fetching chatters failed: {resp.status}', resp.reason)

            all_ = []
            for x in data['chatters'].values():
                all_ += x

            return Chatters(data['chatter_count'], all_, *data['chatters'].values())

    async def create_clip(self, token: str, broadcaster_id: Union[str, int]):
        """|coro|

        Method which creates and returns an edit clip url for the given broadcaster_id.

        Parameters
        ------------
        token: str
            The token to use for Authorization.
        broadcaster_id: Union[str, int]
            The id of the stream you wish to create a clip for.

        Raises
        --------
        HTTPException
            Bad request while creating a clip.
        """

        return await self.http.create_clip(token, broadcaster_id)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2019 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import inspect
from typing import Union

from .errors import *


class Command:

    def __init__(self, name: str, func, **attrs):
        if not inspect.iscoroutinefunction(func):
            raise TypeError('Command callback must be a coroutine.')

        self._name = name
        self._callback = func

        self._checks = []

        try:
            self._checks.extend(func.__checks__)
        except AttributeError:
            pass

        self.aliases = attrs.get('aliases', None)
        sig = inspect.signature(func)
        self.params = sig.parameters.copy()

        self.on_error = None
        self._before_invoke = None
        self._after_invoke = None
        self.no_global_checks = attrs.get('no_global_checks', False)

        self.instance = None

        for key, value in self.params.items():
            if isinstance(value.annotation, str):
                self.params[key] = value.replace(annotation=eval(value.annotation, func.__globals__))

    @property
    def name(self):
        """Returns the name of the command."""
        return self._name

    async def _convert_types(self, param, parsed):

        converter = param.annotation
        if converter is param.empty:
            if param.default in (param.empty, None):
                converter = str
            else:
                converter = type(param.default)

        try:
            argument = converter(parsed)
        except Exception:
            raise BadArgument(f'Invalid argument parsed at `{param.name}` in command `{self.name}`.'
                              f' Expected type {converter} got {type(parsed)}.')

        return argument

    async def parse_args(self, instance, parsed):
        iterator = iter(self.params.items())
        index = 0
        args = []
        kwargs = {}

        try:
            next(iterator)
            if instance:
                next(iterator)
        except StopIteration:
            raise CommandError(f'self or ctx is a required argument which is missing.')

        for _, param in iterator:
            index += 1
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                try:
                    argument = parsed.pop(index)
                except (KeyError, IndexError):
                    if param.default is param.empty:
                        raise MissingRequiredArgument(param)
                    args.append(param.default)
                else:
                    argument = await self._convert_types(param, argument)
                    args.append(argument)

            elif param.kind == param.KEYWORD_ONLY:
                rest = ' '.join(parsed.values())
                if rest.startswith(' '):
                    rest = rest.lstrip(' ')

                if rest:
                    rest = await self._convert_types(param, rest)
                elif param.default is param.empty:
                    raise MissingRequiredArgument(param)
                else:
                    rest = param.default

                kwargs[param.name] = rest
                parsed.clear()
                break
            elif param.VAR_POSITIONAL:
                args.extend(parsed.values())
                break

        if parsed:
            pass  # TODO Raise Too Many Arguments.

        return args, kwargs

    def error(self, func):
        """Decorator which registers a coroutine as a local error handler.

        event_command_error will still be invoked alongside this local handler.
        The local handler must take ctx and error as parameters.

        Parameters
        ------------
        func : :ref:`coroutine <coroutine>`
            The coroutine function to register as a local error handler.

        Raises
        --------
        twitchio.TwitchIOBException
            The func is not a coroutine function.
        """
        if not inspect.iscoroutinefunction(func):
            raise CommandError('Command error handler must be a coroutine.')

        self.on_error = func
        return func

    def before_invoke(self, func):
        """Decorator which registers a coroutine as a before invocation hook.

        The hook will be called before a command is invoked.
        The hook must take ctx as a sole parameter.

        Parameters
        ------------
        func: :ref:`coroutine <coroutine>`
            The coroutine function to register as a before invocation hook.

        Raises
        --------
        twitchio.TwitchIOBException
            The func is not a coroutine function.
        """
        if not inspect.iscoroutinefunction(func):
            raise CommandError('Before invoke func must be a coroutine')

        self._before_invoke = func
        return func

    def after_invoke(self, func):
        """Decorator which registers a coroutine as a after invocation hook.

        The hook will be called after a command is invoked.
        The hook must take ctx as a sole parameter.

        Parameters
        ------------
        func: :ref:`coroutine <coroutine>`
            The coroutine function to register as an after invocation hook.

        Raises
        --------
        twitchio.TwitchIOBException
            The func is not a coroutine function.
        """
        if not inspect.iscoroutinefunction(func):
            raise CommandError('After invoke func must be a coroutine.')

        self._after_invoke = func
        return func


def command(*, name: str=None, aliases: Union[list, tuple]=None, cls=None, no_global_checks=False):
    """Decorator that turns a coroutine into a Command.

    Parameters
    ------------
    name : Optional[str]
        The name of the command.
    aliases : Optional[Union[list, tuple]]
        The aliases of the command.
    cls : Optional[class]
        The class to build the command from. This must be compatible as a Command E.g A subclass.

    Raises
    --------
    TypeError
        cls was not type class.
    """
    if cls and not inspect.isclass(cls):
        raise TypeError(f'cls must be of type <class> not <{type(cls)}>')

    cls = cls or Command

    def decorator(func):
        fname = name or func.__name__
        command = cls(name=fname, func=func, aliases=aliases, no_global_checks=no_global_checks)

        return command
    return decorator


def check(predicate):
    """A decorator that adds a check to a command.

    Parameters
    ------------
    predicate : bool
        The predicate to check. This must return True, False or raise. A truth value means the check has passed.
    """
    # todo Better docstrings/examples
    def decorator(func):
        if isinstance(func, Command):
            func._checks.append(predicate)
        elif not hasattr(func, '__checks__'):
            func.__checks__ = [predicate]
        else:
            func.__checks__.append(predicate)
        return func
    return decorator


class AutoCog:
    pass


def cog(*, name: str=None, **attrs):
    """A decorator that turns a class into a Cog.

    Parameters
    ------------
    name : Optional[str]
        The name of the cog. If no name was passed, the class name will be used in place.
    \*\*attrs:
        Extra attributes to set to the class, as kwargs.
    """
    def wrapper(klass):
        class Cog(AutoCog, klass):

            def __new__(cls, *args, **kwargs):
                result = super().__new__(cls)

                result.__module__ = klass.__module__
                result.__name__ = name or klass.__name__

                for k, v in attrs.items():
                    setattr(result, k, v)

                return result

            def _prepare(self, bot):
                for name_, member in inspect.getmembers(self):
                    if isinstance(member, Command):
                        member.instance = self
                        bot.add_command(member)

                    elif name_.startswith('event_'):
                        bot.add_listener(member, name_)

                bot.cogs[self.__name__] = self

        return Cog
    return wrapper



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-2019 TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

__all__ = ('Message', 'Channel', 'User', 'Context', 'NoticeSubscription')


import datetime
import time
from typing import *

from .abcs import Messageable
from .errors import EchoMessageWarning


class Message:

    __slots__ = ('_author', '_channel', '_raw_data', 'content', 'clean_content', '_tags', '_timestamp', 'echo')

    def __init__(self, **attrs):
        self._author = attrs.pop('author', None)
        self._channel = attrs.pop('channel', None)
        self._raw_data = attrs.pop('raw_data', None)
        self.content = attrs.pop('content', None)
        self.clean_content = attrs.pop('clean_content', None)
        self._tags = attrs.pop('tags', None)
        self.echo = False

        try:
            self._timestamp = self._tags['tmi-sent-ts']
        except (TypeError, KeyError):
            self._timestamp = time.time()

    @property
    def author(self) -> 'User':  # stub
        """The User object associated with the Message."""
        return self._author

    @property
    def channel(self) -> 'Channel':  # stub
        """The Channel object associated with the Message."""
        return self._channel

    @property
    def raw_data(self) -> str:
        """The raw data received from Twitch for this Message."""
        return self._raw_data

    @property
    def tags(self) -> Optional[dict]:
        """The tags associated with the Message.

        Could be None.
        """
        return self._tags

    @property
    def timestamp(self) -> datetime.datetime.timestamp:
        """The Twitch timestamp for this Message.

        Returns
        ---------
        timestamp:
            UTC datetime object of the Twitch timestamp.
        """
        timestamp = datetime.datetime.utcfromtimestamp(int(self._timestamp) / 1000)
        return timestamp


class Channel(Messageable):

    __slots__ = ('_channel', '_ws', '_http', '_echo', '_users')

    def __init__(self, name, ws, http):
        self._channel = name
        self._http = http
        self._ws = ws
        self._echo = False
        self._users = {}

    def __str__(self):
        return self._channel

    @property
    def name(self) -> str:
        """The channel name."""
        return self._channel

    @property
    def chatters(self) -> list:
        """The channel's chatters."""
        return list(self._users.values())

    def _get_channel(self) -> Tuple[str, None]:
        return self.name, None

    def _get_method(self) -> str:
        return self.__class__.__name__

    @property
    def _get_socket(self):  # stub
        if self._echo is True:
            raise EchoMessageWarning('Unable to respond to Echo-Messages.')

        return self._ws

    async def get_stream(self) -> dict:
        """|coro|

        Method which retrieves stream information on the channel, provided it is active (Live).

        Returns
        ---------
        dict:
            Dict containing active streamer data. Could be None if the stream is not live.

        Raises
        --------
        HTTPException
            Bad request while fetching streams.
        """

        data = await self._http.get_streams(channels=[self.name])

        try:
            return data[0]
        except IndexError:
            pass


class User:

    __slots__ = ('_name', '_channel', '_tags', 'display_name', '_id', 'type',
                 '_colour', 'subscriber', 'turbo', '_badges', '_ws', '_mod')

    def __init__(self, ws, **attrs):
        self._name = attrs.pop('author', None)
        self._channel = attrs.pop('channel', self._name)
        self._tags = attrs.pop('tags', None)
        self._ws = ws

        if not self._tags:
            self._tags = {}

        self.display_name = self._tags.get('display-name', self._name)
        self._id = int(self._tags.get('user-id', 0))
        self.type = self._tags.get('user-type', 'Empty')
        self._colour = self._tags.get('color', None)
        self.subscriber = self._tags.get('subscriber', None)
        self.turbo = self._tags.get('turbo', None)

        self._badges = {}
        badges = self._tags.get('badges', None)
        if badges:
            for chunk in badges.split(','):
                k, _, v = chunk.partition('/')
                self._badges[k] = int(v)

        self._mod = self._tags.get('mod', 0) if self._tags else attrs.get('mod', 0)

    def __repr__(self):
        return '<User name={0.name} channel={0._channel}>'.format(self)

    def __eq__(self, other):
        return other == self.name

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        """The user's name."""
        return self._name

    @property
    def id(self) -> int:
        """The user's ID.

         Could be 0 if no Tags were received."""
        return self._id

    @property
    def channel(self) -> Channel:
        """The channel object associated with the User.

        .. note::

            The channel will be valid for the data which triggered the Event. It is possible the
            user could be in multiple channels. E.g: The User BobRoss sends a message from the Channel ArtIsCool.
            The Channel object received will be ArtIsCool.
        """
        return self._channel

    @property
    def colour(self) -> Optional[str]:
        """The user's colour.

        Could be None if no Tags were received.
        """
        return self._colour

    @property
    def color(self) -> Optional[str]:
        """An American-English alias to colour."""
        return self.colour

    @property
    def is_turbo(self) -> bool:
        """A boolean indicating whether the User is Turbo.

        Could be None if no Tags were received.
        """
        return self.turbo

    @property
    def is_subscriber(self) -> bool:
        """A boolean indicating whether the User is a subscriber of the current channel.

        Could be None if no Tags were received.
        """
        return self.subscriber

    @property
    def badges(self) -> dict:
        """The badges associated with the User.

        Could be an empty Dict if no Tags were received.
        """
        return self._badges

    @property
    def tags(self) -> dict:
        """The Tags received for the User.

        Could be an empty Dict if no tags were received.
        """
        return self._tags

    @property
    def is_mod(self) -> bool:
        """A boolean indicating whether the User is a moderator of the current channel."""
        if self._mod == 1:
            return True
        if self.channel.name == self.display_name.lower():
            return True
        else:
            return False


class Context(Messageable):
    """
    The context of which a command is invoked under.

    Attributes
    ------------
    author : :class:`.User`
        The author of the command.
    prefix : str
        The prefix associated with the command.
    message : :class:`.Message`
        The message associated with the command.
    channel : :class:`.Channel`
        The channel associated with the command.
    command : :class:`twitchio.ext.core.Command`
        The command which was invoked.


    .. note::
        Context is only available through the commands extension.
    """

    def __init__(self, message: Message, channel: Channel, user: User, **attrs):
        self.message = message
        self.channel = channel
        self.content = message.content
        self.author = user
        self.prefix = attrs.get('prefix', None)

        self._echo = self.channel._echo
        self._ws = self.channel._ws

        self.command = attrs.get('Command', None)
        self.args = attrs.get('args', None)
        self.kwargs = attrs.get('kwargs', None)

    def _get_channel(self) -> Tuple[str, None]:
        return self.channel.name, None

    def _get_method(self) -> str:
        return self.__class__.__name__

    @property
    def _get_socket(self):  # stub
        if self._echo is True:
            raise EchoMessageWarning('Unable to respond to Echo-Messages.')

        return self._ws

    async def get_stream(self) -> dict:
        """|coro|

        Method which retrieves stream information on the channel stored in Context, provided it is active (Live
        Returns
        ---------
        dict
            Dict containing active streamer data. Could be None if the stream is not live.

        Raises
        --------
        HTTPException
            Bad request while fetching streams.
        """
        return await self.channel.get_stream()


class NoticeSubscription:
    """
    The Dataclass sent to `event_usernotice_subscription` events.

    Attributes
    ------------
    channel : :class:`.Channel`
        The channel associated with the subscription event.
    user : :class:`.User`
        The user associated with the subscription event.
    tags : dict
        The raw tags dict associated with the subscription event.
    cumulative_months : Optional[int]
        The total number of months the user has subscribed. Could be None if not provided.
    share_streak : bool
        Boolean indicating whether users want their streaks to be shared.
    streak_months : Optional[int]
        The number of consecutive months the user has subscribed. Could be None if not provided.
        This is 0 if ``share_streak`` is False.
    sub_plan : str
        The type of subscription plan being used.
        Valid values: Prime, 1000, 2000, 3000.

        1000, 2000, and 3000 refer to the first, second, and third levels of paid subscriptions,
        respectively (currently $4.99, $9.99, and $24.99).
    sub_plan_name : str
        The display name of the subscription plan.
        This may be a default name or one created by the channel owner.
    """

    def __init__(self, *, channel: Channel, user: User, tags: dict):
        self.channel = channel
        self.user = user

        self.tags = tags

        self.cumulative_months = tags.get('msg-param-cumulative-months', None)
        if self.cumulative_months:
            self.cumulative_months = int(self.cumulative_months)

        self.share_streak = bool(tags['msg-param-should-share-streak'])

        self.streak_months = tags.get('msg-param-streak-months', None)
        if self.streak_months:
            self.streak_months = int(self.streak_months)

        self.sub_plan = tags['msg-param-sub-plan']
        self.sub_plan_name = tags['msg-param-sub-plan-name']
